{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE OverloadedStrings #-}
module Main where

import Control.Applicative

import System.Environment
import System.Exit
import System.IO

import Text.Parsec.Pos

import Types
import Parser
import Eval
import Control.Exception

import Data.Maybe
import Data.Either

import Control.Monad (forM_, forever)
import Control.Concurrent (MVar, newMVar, modifyMVar_, modifyMVar, readMVar, swapMVar)
import Data.Text (Text, unpack)
import Data.Text.Encoding

import qualified Data.ByteString.Lazy.Char8 as TLC
import qualified Data.ByteString.Lazy as T
import qualified Data.ByteString.Lazy as TL

import qualified Data.Text.Lazy.Encoding as T

import qualified Network.WebSockets as WS
import qualified Data.Text as T
import qualified Data.Text.IO as T

import Data.Aeson

errorJSON :: String -> Value
errorJSON s = object [
    "error" .= toJSON s
  ]

data Result =
  Err String

data Cmd =
    Load { load :: String }
  |
    Step { step :: Int }
  |
    In   { inp :: Int }
  |
    Nop
    deriving Show

instance FromJSON Cmd where
  parseJSON v = parseLoad v <|> parseStep v <|> parseIn v
    where
      parseLoad = withObject "program" $ \o -> Main.Load <$> o .: "load"
      parseStep = withObject "step" $ \o -> Main.Step <$> o .: "step"
      parseIn   = withObject "input" $ \o -> Main.In <$> o .: "input"

------------------------------------------------------------------------------

type Client = (Text, WS.Connection)
type ServerState = [(ST, [(Command,Line)])]

newServerState :: ServerState
newServerState = []

------------------------------------------------------------------------------

main :: IO ()
main = do state <- newMVar newServerState
          putStrLn "Waiting on ws://127.0.0.1:9160"
          WS.runServer "127.0.0.1" 9160 $ application state

application :: MVar ServerState -> WS.ServerApp
application state pending = do
  conn <- WS.acceptRequest pending
  print "connected"
  talk ("br-gaster", conn) state

talk :: Client -> MVar ServerState -> IO ()
talk (user, conn) state = forever $ do
  (msg :: TL.ByteString) <- WS.receiveData conn
  TLC.putStrLn msg
  prog <- case decode msg of
    -- load program and jump to main
    Just (Main.Load prog) -> do
      case readCommands prog of
        Left err -> WS.sendTextData conn (encode $ errorJSON err)
        Right p -> do
          e  <- begin p
          case e of
            Left err -> putStrLn err
            Right (s,cmds) -> do
              swapMVar state [(s,cmds)]
              let line = if length cmds /= 0 then snd $ head cmds else -1
              WS.sendTextData conn (encode $ runstackJSON s line)
              --putStrLn $ show (s, cmds)
      return ()
    -- step
    -- TODO: add check that program was previously loaded!
    Just (Main.Step i) -> do
      [(s, cmds)] <- readMVar state
      let line = if length cmds /= 0 then snd $ head cmds else -1
      if outstandingInput s
        then WS.sendTextData conn (encode $ errorJSON "expecting input")
        else do
          e  <- Eval.step (clearOutputs s) cmds
          case e of
            Left err -> WS.sendTextData conn (encode $ errorJSON err)
            Right (s,cmds) -> do
              swapMVar state [(s,cmds)]
              WS.sendTextData conn (encode $ runstackJSON s line)
              --putStrLn $ show (s, cmds)
      return ()

    Just (Main.In i) -> do
      [(s, cmds)] <- readMVar state
      if outstandingInput s
        then do
          let s' = clearInputs $ pushInput s i
          swapMVar state [(s',cmds)]
          let line = if length cmds /= 0 then snd $ head cmds else -1
          WS.sendTextData conn (encode $ runstackJSON s' line)
        else WS.sendTextData conn (encode $ errorJSON "unexpeccted input")
      return ()

    Just cmd -> return ()
    Nothing  -> return ()
  -- WS.sendTextData conn prog
  return () --putStrLn $ show prog
